
#include "..\..\common\ImageFilter.h"


using namespace FilterSpace;
using namespace System::Windows::Forms;

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);

	int PixelSizeIn = f->pixelsizein / 8;
	int PixelSizeOut = f->pixelsizeout / 8;
	int ChannelsIn = f->channelcountin;
	int ChannelsOut = f->channelcountout;
	int StrideIn = f->stridein;
	int StrideOut = f->strideout;
	byte *DataIn = f->datain;
	byte *DataOut = f->dataout;
	byte *mask = f->mask;

	int x, y;
	byte *din, *din0 = DataIn + sy * StrideIn + sx * PixelSizeIn;
	byte *dout, *dout0 = DataOut;
	int mstride = (f->width + 7) >> 3;
	int processed_pixels = 0;

//	MessageBox::Show((mask == 0).ToString() + "\n" + PixelSizeIn.ToString() + " -> " + PixelSizeOut.ToString() + "\n" + ChannelsIn.ToString() + " -> " + ChannelsOut.ToString() + "\n" + StrideIn.ToString() + " -> " + StrideOut.ToString());

	int value;
	int maxch = (ChannelsIn > ChannelsOut) ? ChannelsOut : ChannelsIn;

	for (y = 0; y < sh; ++y)
	{
		din = din0;
		dout = dout0;

		for (x = 0; x < sw; ++x)
		{
			bool process = !mask;
			if (mask) 
			{
				int z = x + sx;
				process = (mask[z >> 3] & ((z & 0x07) ? (1 << (z & 0x07)) : 1)) != 0;
			}

			if (process)
			{
				if (ChannelsIn == 3)
					value = (byte)(din[0] * 0.114 + din[1] * 0.587 + din[2] * 0.299);
				else 
				{
					value = 0;
					for (int t = 0; t < ChannelsIn; ++t) value += din[t];
					value /= ChannelsIn;
				}
				for (int t = 0; t < ChannelsOut; ++t) dout[t] = value;
			}
			else
			{
				for (int t = 0; t < maxch; ++t) dout[t] = din[t];
			}
			din += PixelSizeIn;
			dout += PixelSizeOut;
		}

		if (mask) mask += mstride;
		din0 += StrideIn;
		dout0 += StrideOut;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;

		if (updater != nullptr && processed_pixels > 10000) // MULTI-THREAD
		{
			if (updater->WaitOne(10, false))
			{
				*(f->processing_ptr) += processed_pixels;
				processed_pixels = 0;
				updater->Release(1);
			}
		}
		else // SINGLE-THREAD
		{
			if (f->processing_ptr) *(f->processing_ptr) += processed_pixels;
			processed_pixels = 0;
		}
		
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	
	return true;
}


//
//
IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, width = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);
	xml_node^ xn = gcnew xml_node( fs );

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* f = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(f);
	delete [] f;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			xt->XmlGetChildValueByName("Stride", stride);
			xt->XmlGetChildValueByName("Height", height);
			if (stride * height == 0) return IntPtr(0);

			DialogResult res = MessageBox::Show("Would you like to create an 8bit grayscale image\ninstead of a grayscale image with the original format?", "Output format", MessageBoxButtons::YesNo, MessageBoxIcon::Question);
			if (res == DialogResult::Yes)
			{
				xt->XmlGetChildValueByName("Width", width);
				xml_node^ xt2;
				stride = ((width + 3) >> 2) << 2;
				xt->XmlSetChildValueByName("Stride", stride);
				xt->XmlSetChildValueByName("ChannelCount", 1);
				xt->XmlSetChildValueByName("PixelSize", 8);
				xt->XmlRemoveChildByName("ChannelsBitAllocation");
				xt->children.Add(xt2 = gcnew xml_node(xt, "ChannelsBitAllocation", "", ""));
				SetChannelsBitAllocation(xt2, 0, 8);

				byte binary_color_map[1024];
				memset(binary_color_map, 0, 1024);
				for (int i = 1; i < 256; i++) binary_color_map[i * 4] = binary_color_map[i * 4 + 1] = binary_color_map[i * 4 + 2] = i;
				String^ colMap = stringToString(base64((unsigned char*)(binary_color_map), 1024));
				xt2->XmlSetChildValueByName("ColorMap", colMap, false);
			}
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	return (IntPtr)(void*)xn->XML_Tree_Marshal();
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) return 0;

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;


	return filset;
}



bool ImageFilter::CanLocalize(){ return true; }

IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,           0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}




// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr) = 0;
		*(filset->processing_size) = filset->width * filset->height;

		int cpucores = System::Environment::ProcessorCount;

		if (!filset->nCoreEnabled || cpucores == 1)
		{
			b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
		}
		else
		{
			filter_core_func f = (::FilterCore);

			if (FILTER_TYPE == IMAGEFILTER_PIXELWISE)
			{
				filset->nCoreExecutionSize = 1; // SAFETY CONDITIONING
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BOX)
			{
				b = ImageFilterBase::ExecuteFilterBox(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BLOCK)
			{
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else // if (FILTER_TYPE == IMAGEFILTER_RECURSIVE) AND OTHERS
			{
				// NO AUTOMATIC MULTITHREAD SUPPORT
				b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
			}
		}
		delete filset;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}


